/*
 *
 * Function for recognizing partitionability
 * (for pure 2-dimensional complexes)
 *
 * by M.Hachimori, Jul 2006.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"
/* #define MAXVERT 100 */

int m_empty;
int m_vert[MAXVERT];
int m_edge[MAXVERT][MAXVERT];
int c_empty;
int c_vert[MAXVERT];
int c_edge[MAXVERT][MAXVERT];
int f_vec[MAXDIM], h_vec[MAXDIM];

void cancel2(Simplices* facets, int num);
int check2(Simplices* facets, int num);
void putpart2(Simplices* facets, int num);
void printdata2(Simplices* facets);

/*
 * Function for recognizing partitionability
 * (for pure 2-dimensional complexes)
 *   Input: list of facets
 *   return 1 if partitionable
 *   return 0 if not partitionable
 */
int is_partitionable2(facets)
  Simplices *facets;
{

  Simplices *faces;
  Simplices *tmp_facets;
  int i, j, dim = 2;

  /* Check the complex is pure 2-dimensional.*/
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    if (tmp_facets->vert[0] != 2) {
      printf("There are facets of dimension != 2. Error.\n");
      return(-1);
    }
    tmp_facets = tmp_facets->next;
  }

  /* get h-vector */
  faces = facets2faces(facets);
  get_fvector(faces, f_vec);
  f_to_h(f_vec, h_vec, facets->vert[0]);

  /* check h-vector is nonnegative */
  for (i = 0; i <= dim+1; i ++)
    if (h_vec[i] < 0) {
      printf("h-vector has negative component. Not partitionable\n");
      return(0);
    }

  /* initialization */
  /* c_xxx indicate whether faces are covered or not */
  for (i = 0; i < MAXVERT; i ++) {
    c_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++)
      c_edge[i][j] = 0;
    }
  c_empty = 0;
  /* m_xxx indicate the multiplicity of incident facets */
  for (i = 0; i < MAXVERT; i ++) {
    m_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++)
      m_edge[i][j] = 0;
    }
  m_empty = 0;
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    m_vert[tmp_facets->vert[1]] ++;
    m_vert[tmp_facets->vert[2]] ++;
    m_vert[tmp_facets->vert[3]] ++;
    m_edge[tmp_facets->vert[1]][tmp_facets->vert[2]] ++;
    m_edge[tmp_facets->vert[1]][tmp_facets->vert[3]] ++;
    m_edge[tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
    m_empty ++;
    tmp_facets->vert[4] = -1;  /* facets->vert[4] indicates the type of interval
                                 in a partition. 
                                   0: [empty,facet]
                                   1,2,3 : [1,facet], [2,facet], [3,facet]
                                   4,5,6 : [12,facet], [13,facet], [23,facet]
                                   7: [facet,facet]
                                   -1: not assigned yet.  */
    tmp_facets = tmp_facets->next;
  }

  /* main loop */
  tmp_facets = facets;
  while(tmp_facets != NULL) {
#ifdef DEBUG
    printdata2(facets);
#endif
    /* check types in the order -1,7,0,1,2,3,4,5,6 */
    if (tmp_facets->vert[4] == 6) {
      tmp_facets->vert[4] = -1;
      if (tmp_facets->prev == NULL) return 0;
      else {
        tmp_facets = tmp_facets->prev;
        cancel2(tmp_facets, tmp_facets->vert[4]);
        continue;
        }
      }
    else {
      if (tmp_facets->vert[4] == -1) tmp_facets->vert[4] = 7;
      else if (tmp_facets->vert[4] == 7) tmp_facets->vert[4] = 0;
      else tmp_facets->vert[4] ++;
      if (check2(tmp_facets, tmp_facets->vert[4])) {
        putpart2(tmp_facets, tmp_facets->vert[4]);
        if (tmp_facets->next == NULL) goto PARTITIONABLE;
        else {
          tmp_facets = tmp_facets->next;
          continue;
          }
        }
      }
    }

  return -1;

  PARTITIONABLE:
  printdata2(facets);
  return 1;

}

void cancel2(facets, num)
  Simplices* facets;
  int num;
{
  m_empty ++;
  m_vert[facets->vert[1]] ++;
  m_vert[facets->vert[2]] ++;
  m_vert[facets->vert[3]] ++;
  m_edge[facets->vert[1]][facets->vert[2]] ++;
  m_edge[facets->vert[1]][facets->vert[3]] ++;
  m_edge[facets->vert[2]][facets->vert[3]] ++;
  switch(num) {
    case 0:
      c_empty --;
      c_vert[facets->vert[1]] --;
      c_vert[facets->vert[2]] --;
      c_vert[facets->vert[3]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      h_vec[0] ++;
      break;
    case 1:
      c_vert[facets->vert[1]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      h_vec[1] ++;
      break;
    case 2:
      c_vert[facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      h_vec[1] ++;
      break;
    case 3:
      c_vert[facets->vert[3]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      h_vec[1] ++;
      break;
    case 4:
      c_edge[facets->vert[1]][facets->vert[2]] --;
      h_vec[2] ++;
      break;
    case 5:
      c_edge[facets->vert[1]][facets->vert[3]] --;
      h_vec[2] ++;
      break;
    case 6:
      c_edge[facets->vert[2]][facets->vert[3]] --;
      h_vec[2] ++;
      break;
    case 7:
      h_vec[3] ++;
      break;
    }

  return;

}

int check2(facets, num)
  Simplices* facets;
  int num;
{

  /* h-vector check */
  switch(num) {
    case 0:
      if (h_vec[0] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 1:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 2:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 3:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 4:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 5:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 6:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 7:
      if (h_vec[3] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    }

  switch(num) {
    case 0:
      if (c_empty == 1 ||
          c_vert[facets->vert[1]] == 1 ||
          c_vert[facets->vert[2]] == 1 ||
          c_vert[facets->vert[3]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 1:
      if (
          c_vert[facets->vert[1]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 2:
      if (
          c_vert[facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 3:
      if (
          c_vert[facets->vert[3]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 4:
      if (
          c_edge[facets->vert[1]][facets->vert[2]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 5:
      if (
          c_edge[facets->vert[1]][facets->vert[3]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 6:
      if (
          c_edge[facets->vert[2]][facets->vert[3]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 7:
      break;
  }

  if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[1]] == 0 && 
      m_vert[facets->vert[1]] == 1 &&
        !(num == 0 || num == 1) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[2]] == 0 && 
      m_vert[facets->vert[2]] == 1 &&
        !(num == 0 || num == 2) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[3]] == 0 && 
      m_vert[facets->vert[3]] == 1 &&
        !(num == 0 || num == 3) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[1]][facets->vert[2]] == 0 &&
      m_edge[facets->vert[1]][facets->vert[2]] == 1 &&
        !(num == 0 || num == 1 || num == 2 || num == 4) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[1]][facets->vert[3]] == 0 &&
      m_edge[facets->vert[1]][facets->vert[3]] == 1 &&
        !(num == 0 || num == 1 || num == 3 || num == 5) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[2]][facets->vert[3]] == 0 &&
      m_edge[facets->vert[2]][facets->vert[3]] == 1 &&
        !(num == 0 || num == 2 || num == 3 || num == 6) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }

  return 1;

}

void putpart2(facets, num)
  Simplices* facets;
  int num;
{
  m_empty --;
  m_vert[facets->vert[1]] --;
  m_vert[facets->vert[2]] --;
  m_vert[facets->vert[3]] --;
  m_edge[facets->vert[1]][facets->vert[2]] --;
  m_edge[facets->vert[1]][facets->vert[3]] --;
  m_edge[facets->vert[2]][facets->vert[3]] --;
  switch(num) {
    case 0:
      c_empty ++;
      c_vert[facets->vert[1]] ++;
      c_vert[facets->vert[2]] ++;
      c_vert[facets->vert[3]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      h_vec[0] --;
      break;
    case 1:
      c_vert[facets->vert[1]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      h_vec[1] --;
      break;
    case 2:
      c_vert[facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      h_vec[1] --;
      break;
    case 3:
      c_vert[facets->vert[3]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      h_vec[1] --;
      break;
    case 4:
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      h_vec[2] --;
      break;
    case 5:
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      h_vec[2] --;
      break;
    case 6:
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      h_vec[2] --;
      break;
    case 7:
      h_vec[3] --;
      break;
    }

  return;

}

void printdata2(facets)
  Simplices* facets;
{
  Simplices* tmp_facets;
  int i;

  tmp_facets = facets;
  while(tmp_facets != NULL) {
    for (i = 1; i <= tmp_facets->vert[0] + 1; i ++)
      printf("%d ", tmp_facets->vert[i]);
    printf(": ");
    switch(tmp_facets->vert[4]) {
      case 0:
        printf("empty");
        break;
      case 1:
        printf("%d", tmp_facets->vert[1]);
        break;
      case 2:
        printf("%d", tmp_facets->vert[2]);
        break;
      case 3:
        printf("%d", tmp_facets->vert[3]);
        break;
      case 4:
        printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[2]);
        break;
      case 5:
        printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[3]);
        break;
      case 6:
        printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[3]);
        break;
      case 7:
        printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3]);
        break;
      }
    printf("\n");
    tmp_facets = tmp_facets->next;
    }
  printf("\n");
  return;
}

